#include "stdafx.h"
#include "IOHandlers.h"

PKDRPCHANDLER g_KDRPCHandler = NULL;
PPARAMETERLESSPROC g_KDRPCShutdown, g_KDRPCInit = NULL;

static HMODULE g_hKdClientDll = NULL;
const int g_IOPortBase = 0x5658;

bool TryLoadKdClientDllFromRegistry()
{
	HKEY hKey;
	if (RegOpenKey(HKEY_LOCAL_MACHINE, _T("SOFTWARE\\SysProgs\\VirtualKD\\VirtualBoxIntegration"), &hKey))
		return false;

	TCHAR tszDllPath[MAX_PATH];
	DWORD dwSize = sizeof(tszDllPath);

	if (RegQueryValueEx(hKey, _T("KDClientDLLPath"), NULL, NULL, (LPBYTE)tszDllPath, &dwSize) || (dwSize > sizeof(tszDllPath)))
		return false;

	RegCloseKey(hKey);

	g_hKdClientDll = LoadLibrary(tszDllPath);
	if (!g_hKdClientDll)
		return NULL;

	g_KDRPCHandler = (PKDRPCHANDLER)GetProcAddress(g_hKdClientDll, "VBoxHandler");
	if (!g_KDRPCHandler)
	{
		FreeLibrary(g_hKdClientDll);
		g_hKdClientDll = NULL;
		return false;
	}

	g_KDRPCInit = (PPARAMETERLESSPROC)GetProcAddress(g_hKdClientDll, "VBoxInit");
	if (!g_KDRPCInit)
	{
		g_KDRPCHandler = NULL;
		FreeLibrary(g_hKdClientDll);
		g_hKdClientDll = NULL;
		return false;
	}

	g_KDRPCShutdown = (PPARAMETERLESSPROC)GetProcAddress(g_hKdClientDll, "VBoxShutdown");

	g_KDRPCInit();
	return true;
}

#define VIRTUALKD_HOOKED_VBOX_DEVICE_NAME	"pci"

static PCPDMDEVREG s_pOriginalDevReg = NULL;

static int RTCALL ConstructOverride(PPDMDEVINS pDevIns, int iInstance, PCFGMNODE pCfgHandle)
{
	if (!s_pOriginalDevReg)
		return VINF_OBJECT_DESTROYED;
	pDevIns->pReg = s_pOriginalDevReg;
	int rc = s_pOriginalDevReg->pfnConstruct(pDevIns, iInstance, pCfgHandle);
	if (!RT_SUCCESS(rc))
		return rc;

	if (TryLoadKdClientDllFromRegistry())
	{
		rc = PDMDevHlpIOPortRegister(pDevIns, g_IOPortBase, 2, NULL,  VirtualKDPortOutHandler, VirtualKDPortInHandler, NULL, NULL, "VirtualKD interface");
		if (!RT_SUCCESS(rc))
			return rc;
	}
	return VINF_SUCCESS;
}

struct DevRegisterCallbacksHook : public PDMDEVREGCB
{
	PPDMDEVREGCB pOriginalCallbacks;

	DevRegisterCallbacksHook(const PDMDEVREGCB &Callbacks) : PDMDEVREGCB(Callbacks)
	{
		pfnRegister = &RegisterOverride;
	}

	static int RTCALL RegisterOverride(PPDMDEVREGCB pCallbacks, PCPDMDEVREG pDevReg)
	{
		if (pDevReg)
		{
			if (!strcmp(VIRTUALKD_HOOKED_VBOX_DEVICE_NAME, pDevReg->szName))
			{
				PPDMDEVREG pReg = new PDMDEVREG();
				*pReg = *pDevReg;
				s_pOriginalDevReg = pDevReg;
				pReg->pfnConstruct = ConstructOverride;
				pDevReg = pReg;
			}			
		}
		return (static_cast<DevRegisterCallbacksHook *>(pCallbacks))->pOriginalCallbacks->pfnRegister((static_cast<DevRegisterCallbacksHook *>(pCallbacks))->pOriginalCallbacks, pDevReg);
	}

};

extern "C"
{
	typedef int (*PVBoxDevicesRegisterOriginal)(PPDMDEVREGCB pCallbacks, uint32_t u32Version);
	static int VBoxDevicesRegisterOriginal(PPDMDEVREGCB pCallbacks, uint32_t u32Version)
	{
		PVBoxDevicesRegisterOriginal p = (PVBoxDevicesRegisterOriginal)GetProcAddress(GetModuleHandle(_T("VBoxDD0.dll")), "VBoxDevicesRegister");
		if (!p)
			return VERR_UNRESOLVED_ERROR;
		return p(pCallbacks, u32Version);
	}

	int _VBoxDevicesRegister(PPDMDEVREGCB pCallbacks, uint32_t u32Version)
	{
		if (!pCallbacks)
			return VBoxDevicesRegisterOriginal(pCallbacks, u32Version);
		
		DevRegisterCallbacksHook Callbacks(*pCallbacks);
		Callbacks.pOriginalCallbacks = pCallbacks;
		return VBoxDevicesRegisterOriginal(&Callbacks, u32Version);
	}
}